/* The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Eric Hodel's <drbrain@segment7.net> code.
 * 
 * The Initial Developer of the Original Code is Eric Hodel.
 * Portions created by Eric Hodel are Copyright (C) 2001 Eric
 * Hodel. All Rights Reserved.
 * 
 * Other Contributors:
 *      Christopher Hoess <choess@force.stwing.upenn.edu>
 *      Tim Taylor <tim@tool-man.org>
 */


function LanguageDictionary() {
  this.dictionary = null;

  this.lookupLanguageName = 
  function(languageCode) {
    if (this.getDictionary()[languageCode])
      return this.getDictionary()[languageCode];
    else
      return "";
  }

  this.getDictionary = function() {
    if (this.dictionary == null)
      this.dictionary = createDictionary();

    return this.dictionary;
  }

  function createDictionary() {
    var dictionary = new Array();

    var e = getLanguageNames().getSimpleEnumeration();
    while (e.hasMoreElements()) {
      var property = e.getNext();
      property = property.QueryInterface(
          Components.interfaces.nsIPropertyElement);
      dictionary[property.getKey()] = property.getValue();
    }

    return dictionary;
  }

  function getLanguageNames() {
    return srGetStrBundle(
        "chrome://global/locale/languageNames.properties");
  }
}
const languageDictionary = new LanguageDictionary();


function LinkToolbarHandler() {
  this.items = new Array();
  this.defaultItem = null;

  this.handleLinks = function(nodeList) {

    for (var i = 0; i < nodeList.length; i++) {
      if (nodeList.item(i).rel || nodeList.item(i).rev) 
        linkToolbarHandler.handle(nodeList.item(i));
    }
    ddump("LinkToolbarHandler::handleLinks: handled " + nodeList.length + " links");
  }

  this.handle = function(linkElement) {
    convertRevMade(linkElement);

    if (isExcludedLink(linkElement)) return;

    var relAttributes = linkElement.rel.split(" ");
    for (var i = 0; i < relAttributes.length; i++) {
      var linkType = getLinkType(relAttributes[i]);
      this.getItemForLinkType(linkType).addLink(linkElement);
    }
  }

  this.getItemForLinkType = function(linkType) {
    if (!this.items[linkType])
      this.items[linkType] = createItemForLinkType(linkType);

    return this.items[linkType];
  }

  function createItemForLinkType(linkType) {
    ddump("LinkToolbarHandler::createItemForLinkType: type = " + linkType);

    if (!document.getElementById("link-" + linkType))
      return new LinkToolbarTransientMenu(linkType);

    // XXX: replace switch with polymorphism
    switch(document.getElementById("link-" + linkType).tagName) {
      case "button":
        return new LinkToolbarButton(linkType);

      case "menuitem":
        return new LinkToolbarMenuItem(linkType);

      case "menu":
        return new LinkToolbarMenu(linkType);

      default:
        return new LinkToolbarTransientMenu(linkType);
    }
  }

  this.clear = function() {
    for (var linkType in this.items)
      this.items[linkType].clear();
  }

  function convertRevMade(linkElement) {
    if (!linkElement.rel && linkElement.rev) {
      if (linkElement.rev.toLowerCase() == "made")
        linkElement.rel = linkElement.rev;
    }
  }

  function isExcludedLink(linkElement) {
    if (!linkElement.rel) return true;
    if (isStylesheetLink(linkElement)) return true;

    return false;
  }

  function isStylesheetLink(linkElement) {
    if (linkElement.rel.toLowerCase().search("stylesheet") != -1)
      return true;
    else
      return false;
  }

  function getLinkType(relAttribute) {
    switch (relAttribute.toLowerCase()) {
      case "home":
      case "start":
      case "top":
      case "origin":
        return "top";

      case "up":
      case "parent":
        return "up";

      case "begin":
      case "first":
        return "first";

      case "next":
      case "child":
        return "next";

      case "prev":
      case "previous":
        return "prev";

      case "end":
      case "last":
        return "last";

      case "author":
      case "made":
        return "author";

      case "contents":
      case "toc":
        return "toc";

      default:
        return relAttribute.toLowerCase();
    }
  }
}
const linkToolbarHandler = new LinkToolbarHandler();


function LinkToolbarItem (linkType) {
  this.linkType = linkType;
  this.xulElementId = "link-" + linkType;

  this.getXULElement = function() {
    return document.getElementById(this.xulElementId);
  }

  this.clear = function() {
    this.getXULElement().setAttribute("disabled", "true");
    this.getXULElement().removeAttribute("href");
  }

  this.addLink = function(linkElement) {
    if (this.getXULElement().hasAttribute("href")) return false;

    this.setItem(linkElement);
    return true;
  }

  this.setItem = function(linkElement) {
    this.getXULElement().setAttribute("href", linkElement.href);
    this.getXULElement().removeAttribute("disabled");
  }

  makeTooltip = function(linkElement) {
    var title = "";
    // XXX: don't prepend language name if a title is provided
    if (linkElement.hreflang)
      title += languageDictionary.lookupLanguageName(linkElement.hreflang)
            + " - ";
    if (linkElement.title)
      title += linkElement.title;
    else
      title += getNodeAsText(linkElement);

    // XXX: don't append media if title is provided
    if (linkElement.media)
      title += " - " + linkElement.media;

    return title;
  }

  getNodeAsText = function(node) {
    return condenseWhitespace(getNodeAsTextRecursive(node));
  }

  getNodeAsTextRecursive = function(node) {
    var text = "";
    node.normalize();
    if (node.hasChildNodes()) {
      for (var i = 0; i < node.childNodes.length; i++) {
        if (node.childNodes.item(i).nodeType == Node.TEXT_NODE)
          text += node.childNodes.item(i).nodeValue;
        else if (node.childNodes.item(i).nodeType == Node.ELEMENT_NODE)
          text += getNodeAsTextRecursive(node.childNodes.item(i));
      }
    }

    return text;
  }

  condenseWhitespace = function(text) {
    return text.replace(/\W*$/, "").replace(/^\W*/, "").replace(/\W+/g, " ");
  }
}


function LinkToolbarButton (linkType) {
  this.constructor(linkType);

  this.clear = function() {
    this.__proto__.clear.apply(this);

    this.getXULElement().removeAttribute("tooltiptext");
  }

  this.setItem = function(linkElement) {
    this.__proto__.setItem.apply(this, [linkElement]);

    this.getXULElement().setAttribute("tooltiptext", 
        makeTooltip(linkElement));
  }
}
LinkToolbarButton.prototype = new LinkToolbarItem;


function LinkToolbarMenuItem (linkType) {
  this.constructor(linkType);

  this.addLink = function(linkElement) {
    if (!this.__proto__.addLink.apply(this, [linkElement])) return false;

    enableMoreMenu();
    return true;
  }
}
LinkToolbarMenuItem.prototype = new LinkToolbarItem;


function LinkToolbarMenu (linkType) {
  this.constructor(linkType);

  this.knownLinks = new Array();

  this.clear = function() {
    this.__proto__.clear.apply(this);

    // XXX: don't remove the popup, just it's children.  Then put the
    //      popup in the XUL and eliminate the popup checks and popup 
    //      creation code
    if (this.getPopup())
      this.getXULElement().removeChild(this.getPopup());
    this.knownLinks = new Array();
  }

  this.getPopup = function() {
    return document.getElementById(this.xulElementId + "-menupopup");
  }

  this.addLink = function(linkElement) {
    if (this.isAlreadyAdded(linkElement)) return false;

    this.knownLinks[linkElement.href] = true;
    this.addMenuItem(linkElement);
    this.getXULElement().removeAttribute("disabled");
    enableMoreMenu();
    return true;
  }

  this.isAlreadyAdded = function(linkElement) {
    if (!this.knownLinks[linkElement.href])
      return false;
    if (!match(this.knownLinks[linkElement.href].media, linkElement.media))
      return false;
    if (!match(this.knownLinks[linkElement.href].hreflang, linkElement.hreflang))
      return false;

    return true;
  }
  
  function match(first, second) {
    if (!first && !second) return true;
    if (!first || !second) return false;

    return first == second;
  }

  this.addMenuItem = function(linkElement) {
    if (!this.getPopup())
      this.getXULElement().appendChild(this.createPopup());

    this.getPopup().appendChild(this.createMenuItem(linkElement));
  }

  this.createPopup = function() {
    var popup = document.createElement("menupopup");
    popup.setAttribute("id", this.xulElementId + "-menupopup");

    return popup;
  }

  this.createMenuItem = function(linkElement) {
    // XXX: use XUL templates
    var menuitem = document.createElement("menuitem");
    menuitem.setAttribute("label", 
        makeLabel(linkElement, makeTooltip(linkElement)));
    menuitem.setAttribute("href", linkElement.href);
    menuitem.setAttribute("class", "menuitem-iconic bookmark-item");
    menuitem.setAttribute("linkType", 
        "rdf:http://www.w3.org/1999/02/22-rdf-syntax-ns#linkType");

    return menuitem;
  }

  function makeLabel(linkElement, title) {
    return (title == "") ?  linkElement.rel : title;
  }
}
LinkToolbarMenu.prototype = new LinkToolbarItem;


function LinkToolbarTransientMenu (linkType) {
  this.constructor(linkType);

  this.getXULElement = function() {
    if (this.__proto__.getXULElement.apply(this)) 
      return this.__proto__.getXULElement.apply(this);
    else
      return this.createXULElement();
  }

  this.createXULElement = function() {
    ddump("LinkToolbarTransientMenu::createXULElement: this.linkType = " + this.linkType);
    // XXX: use XUL templates
    var menu = document.createElement("menu");
    menu.setAttribute("id", this.xulElementId);
    menu.setAttribute("label", this.linkType);
    menu.setAttribute("disabled", "true");
    menu.setAttribute("class", "menu-iconic bookmark-item");
    menu.setAttribute("container", "true");
    menu.setAttribute("persist", "collapsed");
    menu.setAttribute("visibility", "true");
    menu.setAttribute("type", "rdf:http://www.w3.org/1999/02/22-rdf-syntax-ns#type");

    document.getElementById("more-menupopup").appendChild(menu);

    return menu;
  }

  this.clear = function() {
    // FIXME: we really want to use this:
    // this.hideXULElement();
    // this.__proto__.clear.apply(this);

    // ...but we have to use this, instead:
    this.removeXULElement();

    this.knownLinks = new Array();
  }

  this.hideXULElement = function() {
    /*
     * FIXME: using "hidden" or "collapsed" leads to a crash when you 
     *        open the More menu under certain circumstances:
     *
     *        1. load document with unknown link type
     *        2. reload document
     *        3. click on More menu
     *        4. crash
     *
     *        or sometimes
     *
     *        1. load document with unknown link type
     *        2. open More menu
     *        3. reload document
     *        4. crash
     */
    this.getXULElement().setAttribute("collapsed", "true");
  }

  this.removeXULElement = function() {
    if (this.__proto__.getXULElement.apply(this))
      this.__proto__.getXULElement.apply(this).parentNode.removeChild(
          this.__proto__.getXULElement.apply(this));
  }

  this.addLink = function(linkElement) {
    if(!this.__proto__.addLink.apply(this, [linkElement])) return false;

    showMiscellaneousSeperator();
  }
}
LinkToolbarTransientMenu.prototype = new LinkToolbarMenu;


function toggleLinkToolbar() {
  goToggleToolbar('linktoolbar', 'cmd_viewlinktoolbar');
  if (isLinkToolbarEnabled())
    profile("doRefresh", doRefresh);
  else
    profile("doClear", doClear);
}

/* 
 * refresh is called on every page load 
 *
 * FIXME: refresh isn't called until the entire document has finished loading
 * XXX: optimization: don't refresh when page reloaded if the page hasn't
 *    been modified
 * XXX: optimization: don't refresh when toolbar is compacted
 */
function refresh() {
  if (!isLinkToolbarEnabled())
    return;

  dump("\n");
  profile("doRefresh", doRefresh);
}

function isLinkToolbarEnabled() {
  if (document.getElementById("cmd_viewlinktoolbar").getAttribute("checked") == "true")
    return true;
  else
    return false;
}

function doRefresh() {
  linkToolbarHandler.handleLinks(getLinkElements());

  // FIXME: browser becomes unresponsive during this call...for as
  //    long as 2 or more seconds on heavily linked documents
  linkToolbarHandler.handleLinks(getAnchorElements());
}

function getLinkElements() {
  return getHeadElement().getElementsByTagName("LINK");
}

function getHeadElement() {
  return window._content.document.getElementsByTagName("HEAD").item(0);
}

function getAnchorElements() {
  // XXX: this includes AREA links, too
  return window._content.document.links;
}

function clear() {
  if (!isLinkToolbarEnabled())
    return;

  profile("doClear", doClear);
}

function doClear() {
  disableMoreMenu();
  hideMiscellaneousSeperator();
  linkToolbarHandler.clear();
}

function clicked(event) {
  if (event.target.nodeName == "menubutton") return;
  if (!event.target.getAttribute("href")) return;
  
  loadURI(event.target.getAttribute("href"));
}

function enableMoreMenu() {
  document.getElementById("more").removeAttribute("disabled");
}
function disableMoreMenu() {
  document.getElementById("more").setAttribute("disabled", "true");
}
function showMiscellaneousSeperator() {
  document.getElementById("misc-seperator").removeAttribute("collapsed");
}
function hideMiscellaneousSeperator() {
  document.getElementById("misc-seperator").setAttribute("collapsed", "true");
}

// Debug and testing code for logging and performance testing

function testRefresh() {
  profile("testIterateLinkElements", testIterateLinkElements);
  profile("testIterateAnchorElements", testIterateAnchorElements);
}

function testIterateLinkElements() {
  var document = window._content.document;

  // technique #1
  //var nodeList = document.getElementsByTagName("LINK");

  // technique #2
  var node = document.firstChild;
  while (node.nodeName != "HTML")
    node = node.nextSibling;
  if (node.nodeName != "HEAD") {
    node = node.firstChild;
    while (node.nodeName != "HEAD")
      node = node.nextSibling;
  }
  var nodeList = node.getElementsByTagName("LINK");

  // technique #3
  //var nodeList = document.getElementsByTagName("HEAD").item(0).getElementsByTagName("LINK");

  iterateNodeList(nodeList);
}

function testIterateAnchorElements() {
  var document = window._content.document;

  // technique #1
  //var nodeList = document.getElementsByTagName("A");

  // technique #2
  /*
  var node = document.firstChild.firstChild;  // document -> HTML
  while (node.tagName != "BODY") {
    node = node.nextSibling;
  }
  var nodeList = node.getElementsByTagName("A");
  */

  // technique #3
  var nodeList = document.links;

  iterateNodeList(nodeList);
}

function iterateNodeList(nodeList) {
  for (var i = 0; i < nodeList.length; i++)
    nodeList.item(i);
}

function profile(name, func, args) {
  ddump(name + " begin");
  var start = new Date();

  var result;
  if (args)
    result =  func.apply(this, args);
  else
    result =  func();

  var finish = new Date();
  ddump(name + " end");
  ddump(name + ": completed in " + (finish.getTime() - start.getTime()) + " ms");

  return result;
}

function ddump(msg) {
  dump("LinkToolbar::" + msg + "\n");
}
